-------------------------------------------------------------------------------
-- DELLEREN
-- External CD caller.
-- (C) 2016 Mukunda Johnson (mukunda@mukunda.com)
--
-- See LICENSE-DELLEREN.TXT 
-------------------------------------------------------------------------------

local Main      = DellerenAddon
local L         = Main.Locale
local AceEvent  = LibStub( "AceEvent-3.0" )
 
local COMM_PREFIX      = "DELLEREN"
local PROTOCOL_VERSION = 3

-------------------------------------------------------------------------------		  
function Main:OnInitialize()

	self:RegisterSlashCommands()

	self.update_dummy_frame = CreateFrame("Frame")

	self.Config:CreateDB()
	
	self:SendMessage( "DELLEREN_INIT", {
		version = self.version;
	})
end

-------------------------------------------------------------------------------
function Main:OnEnable() 
	
	self.Status:CacheTalents()
	
	for k, v in pairs( self.modules ) do
		if v.OnStart then v:OnStart() end
	end
	
	self:SendMessage( "DELLEREN_ENABLED", {
		version = self.version;
	})
	
	Main.Ignore:Load()
	 
	self.Config:Apply( true )
	
	self:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED", 
						"OnUnitSpellcastSucceeded" )
	self:RegisterEvent( "GROUP_JOINED", "OnGroupJoined" )
	
	self:RegisterEvent( "PLAYER_TALENT_UPDATE", "OnMyTalentsChanged" )
	--self:RegisterEvent( "GLYPH_ADDED", "OnMyTalentsChanged" )
	self:RegisterEvent( "PLAYER_SPECIALIZATION_CHANGED", 
	                    "OnPlayerSpecializationChanged" )
	self:RegisterEvent( "INSPECT_READY", "OnInspectReady" )
	 
	self:RegisterComm( "DELLEREN" )
	
	self:Print( L["Version:"] .. " " .. self.version )
	
	self:ScheduleTimer( 
		function() 
			if IsInGroup() then
				Main.Status:Send(true) 
			else
				Main.Ignore:Reset()
			end
		end, 10 )
	
	self.update_dummy_frame:SetScript( "OnUpdate", 
							  function() Main:OnFrame() end )
	
	self:ScheduleRepeatingTimer( "CheckPing", 30 )
end

-------------------------------------------------------------------------------
function Main:CheckPing()
	self.Status:CheckPing()
end

-------------------------------------------------------------------------------
function Main:OnGroupJoined()
	self.Status:NewGroup()
end

-------------------------------------------------------------------------------
function Main:OnInspectReady( event, guid )
	self.Inspect:OnInspectReady( guid )
end

-------------------------------------------------------------------------------
function Main:OnMyTalentsChanged()
	self.Status:CacheTalents()
	self.Status:Send()
end

-------------------------------------------------------------------------------
function Main:OnPlayerSpecializationChanged( event, player )
	if player ~= nil and player ~= "player" then
		self.Status:PlayerTalentsChanged( self:UnitFullName(player) )
	end
end

-------------------------------------------------------------------------------
local RAID_UNIT_IDS = {}
for i = 1,40 do RAID_UNIT_IDS["raid" .. i] = i end

-------------------------------------------------------------------------------
local PARTY_UNIT_IDS = {}
for i = 1,10 do PARTY_UNIT_IDS["party" .. i] = i end

-------------------------------------------------------------------------------
function Main:IsPartyOrRaidUnit( unit )
	if (IsInRaid() and RAID_UNIT_IDS[unit])
	        or (not IsInRaid() and PARTY_UNIT_IDS[unit]) then
		return true
	end
	return nil
end

-------------------------------------------------------------------------------
function Main:OnUnitSpellcastSucceeded( event, unitID, spell, rank, 
											lineID, spellID )
	
	-- check if we have a raid# unitid if in a raid, or a party# unitid
	-- if not in a raid, the same spellcast may trigger multiple events
	-- with different unitids
	--

	if self:IsPartyOrRaidUnit( unitID ) then
	
		self.Status:OnSpellUsed( unitID, spellID )
		
		-- 113873 = remove talent 
		-- in legion, this might not be necessary
		-- need confirmation that inspects are automatically updated
		-- when a player changes their stuff.
		if spellID == 113873 then
			self.Status:PlayerTalentsChanged( self:UnitFullName(unitID), 10 )
		end
		
		if self.Query.active and not self.Query.buff and self.Query.requested
		   and UnitGUID( unitID ) == UnitGUID( self.Query.unit ) then
		
			if (not self.Query.item and spellID == self.Query.spell) or
			   (self.Query.item and spell == GetItemSpell( self.Query.spell )) then
				
				Main.Query:Success()
			end
		end
	end
	
	if self.Help.active and not self.Help.buff and unitID == "player" then
	   
		if (not self.Help.item and spellID == self.Help.spell) or
		   (self.Help.item and spell == GetItemSpell( self.Help.spell )) then
		   
			self.Help:Success()
			
		end
	end
end

-------------------------------------------------------------------------------
-- Called when a party or raid member applies a buff or debuff to someone.
--
-- @param spellID     Spell ID of buff.
-- @param source,dest GUID of player who buffed and target
--
function Main:OnAuraApplied( spellID, source, dest ) 

	if self.Query.active and self.Query.buff 
	   and self.Query.requested
	   and source == UnitGUID( self.Query.unit )
	   and ((not self.Query.item and spellID == self.Query.spell) 
	       or (self.Query.item and 
	       GetSpellInfo(spellID) == GetItemSpell(self.Query.spell))) then
		   
		-- hack to fix Vigilance which has a self-buff
		if spellID == 114030 and source == dest then return end
	   
		if dest == UnitGUID( "player" ) then
			
			Main.Query:Success()
			
		else
			
			-- cd was cast on someone else! find another one!
			Main.Query:EndRequest()
		end
	end
	
	if self.Help.active and self.Help.buff 
	   and source == UnitGUID( "player" ) 
	   and ((not self.Help.item and spellID == self.Help.spell)
	       or (self.Help.item 
		      and GetSpellInfo(spellID) == GetItemSpell(self.Help.spell))) then
 
		if dest == UnitGUID( self.Help.unit ) then
			
			self.Help:Success() 
			
		else
			
			self.Help:Fail() 
		end
	end
end

-------------------------------------------------------------------------------
function Main:OnCombatLogEvent( event, ... ) 
	local timestamp,evt,_,sourceGUID,_,_,_,destGUID,_,_,_,spellID = ...
	
	if evt == "SPELL_AURA_APPLIED" or evt == "SPELL_AURA_REFRESH" then
		self:OnAuraApplied( spellID, sourceGUID, destGUID )
		
	end
end

-------------------------------------------------------------------------------
function Main:UnitFullName( unit )
	local n,r = UnitName(unit)
	if r and r ~= "" then
		return n .. '-' .. r
	end
	return n
end
 
-------------------------------------------------------------------------------
function Main:UnlockFrames()
	
	self.unlocked = true
	self.db.profile.locked = false
	DellerenUnlockedNotice:Show()
	self:SendMessage( "DELLEREN_UNLOCKED" ) 
end

-------------------------------------------------------------------------------
function Main:LockFrames()
	if not self.unlocked then return end 
	
	self.unlocked = false
	self.Config.db.profile.locked = true
	
	self:SendMessage( "DELLEREN_LOCKED" )
end
 
-------------------------------------------------------------------------------
-- Returns the squared range to a friendly unit.
--
-- Note that the casting range is slightly longer, since this is the
-- range to their center, and casting extends to their hitboxes
--
-- @param unit unitID of friendly unit.
-- @returns distance squared.
--
local function UnitDistance( unit )
	local y,x = UnitPosition( unit )
	y = y or 999999
	x = x or 999999
	local my,mx = UnitPosition( "player" )
	y,x = y - my, x - mx
	local d = x*x + y*y
	return d
end
 
-------------------------------------------------------------------------------
function Main:UnitNearby( unit )
	return UnitDistance( unit ) < 40*40 and IsItemInRange( 34471, unit )
end
 
-------------------------------------------------------------------------------
function Main:GetSpellReadyTime( spell, item )
	if not item then
		--if IsSpellKnown( spell ) then
		
		-- IsSpellKnown fails for certain talents
		-- workaround http://forums.wowace.com/showthread.php?t=20184
		--
		if GetSpellInfo( GetSpellInfo( spell )) then
			local charges = GetSpellCharges( spell ) 
			if charges ~= nil and charges >= 1 then return 0 end
			
			local start, duration, enable = GetSpellCooldown( spell )
			if start == 0 then return 0 end
			
			-- if there is 1 second left on the cd, then it's ready enough
			-- also uh, account for gcd time which may interfere
			return duration - (GetTime() - start) 
		end
	else
		if GetItemCount( spell ) > 0 then
			local start, duration, enable = GetItemCooldown( spell )
			if start == 0 then return 0 end
			
			return duration - (GetTime() - start)
		end
	end
	
	return nil
end

-------------------------------------------------------------------------------
-- Returns true if we have a cd ready to give to someone, and it isn't
-- already being asked for.
--
-- @param ignore_reserve Don't return false if we have our CD reserved,
--                       (check the cooldown only)
-- @param list           List of spell IDs or item IDs to check. May be a 
--                       single id number too.
-- @param item           true if we are checking for items instead of spells.
--
-- @returns nil if no spells or items are ready, or returns the id of one
--          that is ready.
--
function Main:HasCDReady( ignore_reserve, list, item )
	if not ignore_reserve and (self.Help.active) then
		-- someone is already asking us!
		return nil
	end
	
	if UnitIsDeadOrGhost( "player" ) then 
		-- we are dead
		return nil
	end
	
	if type(list) == "number" then
		list = {list}
	end
	
	-- todo: spell reserves
	
	for k,v in ipairs( list ) do
		local t = Main:GetSpellReadyTime( v, item )
		
		if t ~= nil and t <= 1.6 then return v end 
	end
end

-------------------------------------------------------------------------------
function Main:OnCommReceived( prefix, packed_message, dist, sender )
	if prefix ~= COMM_PREFIX then return end -- discard unwanted messages
	
	local result, msg, data = self:Deserialize( packed_message )
	if result == false then return end -- bad message
	
	if not UnitInParty( sender ) then
		-- malicious message from outside
		return
	end
	
	if data.pv ~= PROTOCOL_VERSION then
	
		-- incompatible protocols!
		if data.pv > PROTOCOL_VERSION then
			if not self.warned_protocol_outdated then
				self.warned_protocol_outdated = true
				self:Print( L["One or more players in your raid are using a newer version of Delleren that isn't compatible with yours."] )
			end
		end
		
		if msg == "STATUS" then
			self.Status:SetIncompatible( sender, data )
		end
		
		return
	end
	
	if sender == UnitName( "player" ) then
		-- ignore mirrored messages, (TODO does this happen?)
		return
	end
	
	if data.tar and UnitGUID("player") ~= data.tar then
		-- crossrealm whisper and we are not the intended target.
		return
	end
	
	
	if msg == "CHECK" then
		-- player is checking if we have a cd ready
		
		local id = self:HasCDReady( false, data.ids, data.item )
		
		if id then
			
			self:RespondReady( sender, data.rid, id )
		end
		
	elseif msg == "READY" then
		
		self.Query:HandleReadyMessage( sender, data )
		
	elseif msg == "GIVE" then
		-- player is asking for a CD
		
		if type(data.id) ~= "number" then return end -- bad message
		
		if not self:HasCDReady( false, data.id, data.item ) then
		
			if not data.item then
				
				self:DeclineCD( sender, data.rid, data.id )
			else
				self:DeclineCD( sender, data.rid )
			end
		else
			self.Help:Start( sender, data.id, data.item, data.buff )
		end
		
	elseif msg == "NO" then
		-- player denied our cd request
		
		if data.rid ~= self.Query.rid then
			return
		end
		
		if data.time then
			self.Status:SetSpellCooldown( sender, self.Query.spell, data.time )
		end
		
		-- end current request and try for another target.
		Main.Query:EndRequest()
		
	elseif msg == "STATUS" then
		
		self.Status:UpdatePlayer( sender, data )
		
	elseif msg == "PING" then
		self.Status:OnPing( sender, data )
	end
end
 
-------------------------------------------------------------------------------
local function CrossesRealm( unit )
	local n,r = UnitName( unit )
	return r ~= nil
end

-------------------------------------------------------------------------------
-- Send a message to other players.
--
-- Handles cross-realm compatibility workarounds.
--
-- @param msg    Message type string
-- @param data   Message data block
-- @param dist   Distribution type.
-- @param target WHISPER distribution target name.
--
function Main:Comm( msg, data, dist, target )
	
	if target ~= nil and dist == "WHISPER" and CrossesRealm( target ) then
		dist = "RAID"
		data.tar = UnitGUID( target )
	end
	
	if dist == "RAID" and IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
	   --and IsInInstance() and not C_Garrison:IsOnGarrisonMap() then -- thisbreaks ashran
	
		dist = "INSTANCE_CHAT"
	end
	
	data.pv = PROTOCOL_VERSION
	
	local packed = self:Serialize( msg, data )
	
	if dist == "WHISPER" then
	
		self:SendCommMessage( COMM_PREFIX, packed, dist, target, "ALERT" )
	else
		self:SendCommMessage( COMM_PREFIX, packed, dist, nil, "ALERT" )
	end
end

-------------------------------------------------------------------------------
-- Send a READY response.
--
-- @param target Name of player to respond to.
-- @param rid    Request ID.
-- @param id     Spell or item ID that we have ready.
--
function Main:RespondReady( target, rid, id )

	local data = {
		rid = rid;
		id  = id;
	}
	
	self:Comm( "READY", data, "WHISPER", target )
end

-------------------------------------------------------------------------------
-- Decline giving a CD, sending a "NO" response
--
-- @param target Name of player we are denying.
-- @param rid    Request ID.
--
function Main:DeclineCD( target, rid, spell )
	local data = { rid = rid }
	
	if spell then
		local time = Main:GetSpellReadyTime( spell, false )
		if time ~= nil and time >= 3 then data.time = time end
	end
	
	self:Comm( "NO", data, "WHISPER", target )
end

-------------------------------------------------------------------------------
-- Frame update handler.
--
function Main:OnFrame()
	self.Status:PeriodicRefresh()
	self.Query:Update()
	self.Help:Update() 
	
	if not self.Query.active and not self.Help.active then
		
		self:DisableFrameUpdates()
	end
	
	if self.db.profile.tracking.enabled then
		self.cd_data = self.Status:BuildCooldownData() 
		self:SendMessage( "DELLEREN_CD_DATA", self.cd_data ) 
	end
end

-------------------------------------------------------------------------------
-- Enable combat log parsing. (this used to be a bit more than that)
--
function Main:EnableFrameUpdates()

	if self.updates_enabled then return end
	self.updates_enabled = true
	
	self:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED", "OnCombatLogEvent" )
end

-------------------------------------------------------------------------------
-- Disable animations and combat log parsing.
--
function Main:DisableFrameUpdates()
	if not self.updates_enabled then return end
	self.updates_enabled = false
	
	self:UnregisterEvent( "COMBAT_LOG_EVENT_UNFILTERED", "OnCombatLogEvent" )
end
  
-------------------------------------------------------------------------------
-- Iterates through unit IDs of your party or raid, excluding the player
--
-- Note that the iterator is not valid across frames.
--
function Main:IteratePlayers()
	local raid   = IsInRaid()
	local index  = 0
	local player = UnitGUID("player")
	
	return function()
		
		while true do
			index = index + 1
			local unit = (raid and "raid" or "party") .. index
			if not UnitExists( unit ) then
				return nil
			end
			
			if UnitGUID( unit ) ~= player then
				return self:UnitFullName(unit)
			end
		end
		
	end
end

-------------------------------------------------------------------------------
function Main:Print( text )
	local prefix = "|cffa7000c<Delleren>|r "
	print( prefix .. text )
end

-------------------------------------------------------------------------------
function Main:RegisterModule( name, tbl, options )
	Main.modules[name] = tbl
	
	options = options or {}
	options.name = options.name or name
	
	if options.db then
		self.Config.db_template.profile[name] = options.db
	end
	
	if options.config_section then
		self.Config.options.args[ name .. "_config" ] = {
			name = L[options.name];
			type = "group";
			args = options.config_section;
		}
		
	end
	
	if options.config_general then
		self.Config:AddGeneral( options.config_general )
		
	end
	
	tbl.db = function() return self.db.profile[name] end
end

-------------------------------------------------------------------------------
function Main:GetModuleConfigOption( mod, name )
	return self.Config.options.args[ mod .. "_config" ].args[name]
end

-------------------------------------------------------------------------------
function Main:GetModuleConfigSection( mod )
	return self.Config.options.args[ mod .. "_config" ].args
end

-------------------------------------------------------------------------------
function Main:ModCall( module, method, ... )
	return self.modules[module][method]( self.modules[module], ... )
end 

-------------------------------------------------------------------------------
function Main:ModMethod( module, method, ... )
	return self.modules[module][method]
end 

-------------------------------------------------------------------------------
function Main:ModStaticCall( module, method, ... )
	return self.modules[module][method]( ... )
end 

-------------------------------------------------------------------------------
function Main:SendMessage( msgname, ... )

	-- pass to modules
	for k,v in pairs( self.modules ) do
		if v["Event_" .. msgname] then
			v["Event_" .. msgname]( v, ... )
		end
	end
	
	-- pass to other addons
	AceEvent:SendMessage( msgname, ... )
end
